/* multivec.hpp */
/* MV 1.3.0
	 Hierachical compact multivector type. Header file giving access to C struct layer..
	 Copyright (c) Ian Bell 2004

*/

#include "mv.hpp"

// -----------------
//   Dimensions
// -----------------
// MV_DIM1 - the maximal number of dimensions supported is defined in mv.hpp



#define MV_DIM0 (5) 						 // Maximal dimension handled at each level ("levelity")
#define MV_NUM_LEVEL (1+((MV_DIM1-1)/MV_DIM0)) // Maximal number of levels 

#define MV_ZERO NULL
#define MV_MAX_CAPACITY ((1<<MV_DIM0)+2)

#define MV_EXP_LEVEL	(4) // Diagnopstics generate warning if level>this


// -----------------------
//   Values of BitFlags
// -----------------------
#define MV_INDEX_MAX ((1<<MV_DIM0)-1)
#define MV_INDEX_NIL (1<<MV_DIM0)


#define MV_MAX_CONJUGATION (1<<MV_NUM_CONJUGATION)
#define MV_CONJUGATION_ALL  (MV_MAX_CONJUGATION-1)

#define MV_CONJUGATION_NONE   		(0)
#define MV_CONJUGATION_NEGATION   (1)
#define MV_CONJUGATION_REVERSE    (2)
#define MV_CONJUGATION_INVOLUTION (4)
#define MV_CONJUGATION_MITIAN 	  (8)
#define MV_CONJUGATION_EXT_NEGATION	  (0x10)
#define MV_CONJUGATION_EXT_MODULATION (0x20)

// #define MV_CONJUGATION_IDENTITY (MV_CONJUGATION_NONE)

#define MV_CONJUGATION_CLIFFORD  (MV_CONJUGATION_INVOLUTION|MV_CONJUGATION_REVERSE) 
#define MV_CONJUGATION_HERMITIAN (MV_CONJUGATION_MITIAN|MV_CONJUGATION_CLIFFORD) 



// ----------------
//    Data Types	 
// ----------------

#if MV_FLOAT

#define realToReal REAL_TO_MYFLOAT
#define RealToreal MYFLOAT_TO_REAL

#define rrMUL(A,B) (A)*(B)
#define rDIVrr(A,B) ((A)/(B))
#define rSQUARE(A) ((A)*(A))

#define Real_OPPOSITESIGN(A,B) ((((A)<0)&&((B)>0))||(((A)>0)&&((B)<0)))

#define Real_TINY FLOAT_TINY
#define Real_0 FLOAT_0
#define Real_2 FLOAT_2
#define Real_RECIP(A) (Real_1 / (A))
#define Real_ABS(A) (fabs(A))
#define Real_SQRT(A) (sqrt(A))
#define Real_SQRT2 FLOAT_SQRT2
#define Real_RECIPSQRT2 FLOAT_RECIPSQRT2
#define Real_RANDOM(A) ((FRAC_RANDOM * (A))/ FRAC_1)
#define Real_INFINITY (112233445566778899.112233)
#define Real_TOLERANCE (0.000001)

#define EQUIVALENT_REALS(A,B) (fabs((A)-(B)) < Real_TOLERANCE)

#define ValidReal(R) (ValidReal0(&(R),__FILE__,__LINE__))
boolean ValidReal0(Real * pr,char * fi,int li);

#endif
					
typedef fint mvaxis; // ranges from 0 to MV_DIM1-1

#if MV_NUM_LEVEL<256
typedef uint8 mvlevel;
#endif


#if MV_DIM0 <= 3
typedef uint8 bladesflag;
#define MV_FLAG_ALL ((uint8)((1<<(1<<MV_DIM0))-1))
#define MV_FLAG_NONE (0)
#define MV_FLAG_VCTR (0x16)
#endif

#if MV_DIM0 == 4
typedef uint32 bladesflag; // could use uint16
#define MV_FLAG_ALL ((1<<(1<<MV_DIM0))-1)
#define MV_FLAG_NONE (0)
#define MV_FLAG_VCTR (0x116)
#endif

#if MV_DIM0 == 5
typedef uint32 bladesflag;
#define MV_FLAG_ALL (0xFFFFFFFF)
#define MV_FLAG_NONE (0)
#define MV_FLAG_VCTR (0x00010116)
#endif

#if MV_DIM0 == 6
typedef uint64 bladesflag;
#define MV_FLAG_ALL uint64_MAX;
#define MV_FLAG_NONE (0L)
#define MV_FLAG_VCTR (0x100010116L)
#endif

#if MV_DIM0 > 6
#error "unsupported"
#endif



#if MV_DIM0<=4

#define MV_WID_HEADER (32)
#define MV_WID_REFCOUNT MV_WID_HEADER - (1<<MV_DIM0)- 4 - MV_DIM0 -1 
#define MV_MAX_REFCNT ((1<<MV_WID_REFCOUNT)-2)

#define MV_OTR_LEVEL (0xF) // Maximal number of levels 

typedef struct 															// DIM0 =	1  2  3  4  5  6
{	unsigned int present	:  (1<<MV_DIM0) ;	  //        2  4  8 16 32 64  bitflags for each possible blade;
	unsigned int level  	:  4 ;					// ;
	unsigned int capacity	:  MV_DIM0 ;				//        1  2  3  4  5  6 capacity (0=>1 cell,1=>2 cells etc);
	unsigned int negsig		:	 1 ;						 //													1=> scalar part of square -ve
	unsigned int freed		:	 1 ;						 //													1=> freed (for diagnostics)
	unsigned int noncompacted	:	 1 ;						 //													1=> non-compacted
	unsigned int refcount	:  MV_WID_REFCOUNT ;
} multivectorheader;

#endif


#if MV_DIM0>=5
typedef struct
{	bladesflag 	present;					// bitflags for each possible blade;
	mvlevel			level			: 8 ;		// 0 for multivector in e12345
	mvcapacity	capacity	: 8 ;		// number of cells in coeff array
	unsigned int negsig		:	1 ;						 //													1=> scalar part of square -ve
	unsigned int freed		:	1 ;
	unsigned int noncompacted	:	1 ;
	unsigned int refcount	:	16 ;
} multivectorheader;	 

#define MV_MAX_REFCNT (0xFF)
#define MV_OTR_LEVEL (0xFF) // Maximal number of levels 

#endif

#define MV_VSR_LEVEL (MV_OTR_LEVEL-1) 



struct multivectorstruct0
{	multivectorheader	h;
	Real	 r[OPENENDED];	// real coefficients of each present blade, in index order
};		 

struct multivectorstruct
{	multivectorheader	h;
	multivector m[OPENENDED];	// lower level multivectorcoefficient of each present blade, in index order
};		 

extern multivector * mv_frame;

extern multivector0 MV_scalar;

extern multivector MV_e0_inv[MV_MAX_GHC];
extern multivector MV_einfi[MV_MAX_GHC];

extern mvgrade BitsSetFbi(fullbladeindex);
//extern bladeindex BitsSetBi(bladeindex); // returns at most (1<<MV_DIM0)
//extern bladeindex BitsSetBf(bladesflag); // returns at most (1<<MV_DIM0)

multivector Reference(multivector);


multivector CopyMultivectorScaled(multivector tgt,multivector src,Real smult);

multivector EnsureWritable(multivector a);

boolean IsZero(multivector a);
boolean IsScalar(multivector a);

int NumBlades(multivector a);
int NumBladesWithCorrection(multivector a);
Real PrincipleBlade(fullbladeindex * fbi,multivector a); 

//multivector Multivector0(bladesflag bf, ...);

Real * MultivectorCoefficient(multivector a, fullbladeindex bi);

multivector FullOneVector(int n);

multivector RandomOneVector(int n);
multivector RandomOneVectorScaled(int n,Real mag,mvconjugation conj=0);
multivector RandomMultivector(void);

void MultivectorFromPointFast(multivector a,posn3D * p);
void MultivectorFromSphereFast(multivector a,posn3D * p,real Rsqrd);

void MultivectorFromSphereFastDbl(multivector a,posn3D * p,real Rsqrd);

multivector MultivectorScalarReal(Real x,mvcapacity cap=1) ;

multivector MultivectorPoint(vector3D * p); 

multivector MultivectorRestrictGrade(multivector tgt, multivector src, int grade);
multivector MultivectorProjection(multivector tgt,multivector a,multivector b,mvproduct pro=MV_GEOMETRIC_PRODUCT);
multivector MultivectorProjectionFast(multivector tgt,multivector a,multivector b,multivector b_inv,mvproduct pro);

multivector Multivector3D1(real x,real y,real z);
multivector Multivector4D1(real x,real y,real z,real t);
multivector Multivector5D1(real x,real y,real z,real t,real u);

multivector MultivectorSphere3D(vector3D * p,real sqdRadius,boolean dual); 

multivector BuildMultivector(int cnt,...);
multivector MultivectorBladeReal(Real,fullbladeindex,mvcapacity cap=1);

multivector0 NewMultivector0(int cap,multivector0 src=NULL);

#if MV_NEWS
#define NewMultivector0(C,S) NewMultivectorCore0(__FILE__,__LINE__,C,S)
#define NewMultivector(C,S,L) NewMultivectorCore(__FILE__,__LINE__,C,S,L)
multivector NewMultivectorCore(char * file,int line, int cap,multivector src=NULL,mvlevel level=0);
multivector0 NewMultivectorCore0(char * file,int line, int cap,multivector0 src=NULL);


#define KillMultivector(P) KillMultivectorCore(P,__FILE__,__LINE__)
#define KillMultivector0(P) KillMultivectorCore0(P,__FILE__,__LINE__)
void KillMultivectorCore(multivector * pmv,char *,int);
void KillMultivectorCore0(multivector0 * pmv,char *,int);

#else

#define NewMultivector0(C,S) NewMultivectorCore0(C,S)
#define NewMultivector(C,S,L) NewMultivectorCore(C,S,L)
multivector NewMultivectorCore(int cap,multivector src=NULL,mvlevel level=0);
multivector0 NewMultivectorCore0(int cap,multivector0 src=NULL);


#define KillMultivector(P) KillMultivectorCore(P)
#define KillMultivector0(P) KillMultivectorCore0(P)
void KillMultivectorCore(multivector * pmv);
void KillMultivectorCore0(multivector0 * pmv);

#endif


//#define NewMultivector(C,S,L) NewMultivector0("fish",12345,C,S,L)

multivector MultivectorConjugation(multivector a, mvconjugation cnj);
void MultivectorConjugate(multivector a, mvconjugation cnj);
void MultivectorScale(multivector a, real lambda);

multivector MultivectorAddMerge(multivector a, multivector b);
multivector MultivectorAdd(multivector tgt, multivector b);
multivector MultivectorAddScaled1(multivector tgt, multivector a, multivector b, Real bmult);
multivector MultivectorAddScaled(multivector tgt, multivector a, multivector b, Real amult, Real bmult);

bladeindex BladeBladeProduct(bladeindex bi,bladeindex bj,mvproduct * ppro,mvlevel level,bladesflag inhibit=MV_FLAG_NONE);
multivector0 MultivectorProduct0(multivector0 tgt, multivector0 a, multivector0 b, mvproduct pr=MV_GEOMETRIC_PRODUCT, bladesflag inhibit=MV_FLAG_NONE);
multivector MultivectorProduct(multivector tgt, multivector a, multivector b, mvproduct pr=MV_GEOMETRIC_PRODUCT, bladesflag inhibit=MV_FLAG_NONE);

multivector MultivectorCommutation(multivector r, multivector a, multivector b, mvproduct product,bladesflag inhibit=MV_FLAG_NONE);

multivector MultivectorNormalise(Real * mult,multivector tgt,multivector src );


Real MultivectorPseudoScalarProduct(multivector a,multivector b,mvproduct pro);
Real MultivectorScalarProduct(multivector a,multivector b,mvproduct pro=MV_ONTO_CONTRACTIVE_PRODUCT);
Real MultivectorSquaredModulus(multivector a);

Real ScalarPart(multivector a); // Faster than using general grade restriction call

mvgradesflag GradesPresent(multivector a);


multivector MultivectorDual(multivector a,int axies,mvproduct product=MV_GEOMETRIC_PRODUCT);

multivector MultivectorInverse(multivector a,mvproduct product=MV_GEOMETRIC_PRODUCT,int attempt=0);

multivector MultivectorTransform(multivector tgt,multivector A,multivector x,mvproduct pro);

void VectorFromMultivectorFast(real * pr,multivector0 a,mvgrade k);
void Multivector0FromVectorfFast(multivector0 a,Real * pr,mvgrade k);
void MultivectorFromVectorfFast(multivector a,Real * pr,mvgrade k);
void MultivectorFromVectorFast(multivector a,real * pr,mvgrade k);


Real MultivectorInvertedMagnitude(multivector a,mvproduct pro=MV_LEFT_CLIFFORD);

mvgrade MaxGrade(multivector a);
mvgrade MultivectorJoin(multivector * join,multivector a,multivector b,boolean usenew=0);
multivector MeetAndJoinGrades(mvgrade * pgr,multivector adb,multivector a,multivector b);


#if MV_VSR_LEVEL
multivector GeometricFactoredProduct(multivector a, multivector b,mvconjugation acnj,mvconjugation bcnj,mvcapacity margin);
multivector OutterFactoredProduct(multivector a, multivector b,mvconjugation acnj,mvconjugation bcnj,mvcapacity margin);
#endif


#if MV_RELEASE
void pf_MV(const multivector mv,fullbladeindex bi,Real r);
#else
void bpf_MV(block_handle * pbh,const multivector mv,int way,fullbladeindex bi=0,Real r=Real_1);
#endif 


boolean CheckValidMV(multivector m);


#if MV_DIM1<=32
void InitMultivectors(mvgrade N=MV_DIM1,fullbladeindex negs=0,fullbladeindex nulls=0,int ghc=false); // defaults to Euclidean
#endif
void ExitMultivectors(void);

// ---------------------------------

extern boolean g_i_neg;	// 1=> pseudoscalar has negative signature

extern multivector MV_e0[MV_MAX_GHC];
extern multivector MV_einf[MV_MAX_GHC];
extern multivector MV_ep[MV_MAX_GHC];
extern multivector MV_em[MV_MAX_GHC];
 
extern multivector MV_basis[MV_DIM1];

//extern cm MV_to_cm(multivector src);
//extern cm MVcm_basis[5];
