/* mv.hpp */

/* MV 1.3.0
	 Encapsulating C++ class for C hierachical multivector type .	Header file for access only to C++ class layer.
	  Copyright (c) Ian C.G.Bell 2004. All rights reserved.
	 For a description of methadologies in this code see www.ibell.co.uk/maths/geoalg.htm
	 For version details see multivector.cpp
*/

#define MV_RELEASE (3)     // release version number
#define MV_RELEASE_LOW (0) // release version number
#define MV_DEBUG (0)

#define MV_NEWS (0) // level of diagnostic output

#define MV_DIM1 (32) // Maximal dimension supported 
#define MV_MAX_GHC (1) // Maximal number of GHC generalisations supported.


// Bitflag constuctors for generally conjugised product specfiers
#define MV_CONJUGATION_SHIFT (8)
#define MV_PRODUCT_SHIFT (0x10)

#define MV_LEFT_INVOLUTION (MV_CONJUGATION_INVOLUTION)
#define MV_LEFT_NEGATION	 (MV_CONJUGATION_NEGATION)
#define MV_LEFT_REVERSE    (MV_CONJUGATION_REVERSE)
#define MV_LEFT_MITIAN  	 (MV_CONJUGATION_MITIAN)

#define MV_LEFT_CLIFFORD (MV_LEFT_REVERSE|MV_LEFT_INVOLUTION)
#define MV_LEFT_HERMITIAN (MV_LEFT_CLIFFORD|MV_LEFT_MITIAN)

#define MV_RIGHT_INVOLUTION (MV_CONJUGATION_INVOLUTION<<MV_CONJUGATION_SHIFT)
#define MV_RIGHT_REVERSE    (MV_CONJUGATION_REVERSE   <<MV_CONJUGATION_SHIFT)
#define MV_RIGHT_MITIAN  		(MV_CONJUGATION_MITIAN <<MV_CONJUGATION_SHIFT)
#define MV_RIGHT_CLIFFORD   (MV_LEFT_CLIFFORD<<MV_CONJUGATION_SHIFT)
#define MV_RIGHT_HERMITIAN  (MV_LEFT_HERMITIAN<<MV_CONJUGATION_SHIFT)


#define MV_GEOMETRIC_PRODUCT 				(0)
#define MV_ONTO_CONTRACTIVE_PRODUCT (0x01<<MV_PRODUCT_SHIFT)
#define MV_BY_CONTRACTIVE_PRODUCT 	(0x02<<MV_PRODUCT_SHIFT)
#define MV_OUTTER_PRODUCT 					(0x04<<MV_PRODUCT_SHIFT)
#define MV_INNER_PRODUCT 						(0x08<<MV_PRODUCT_SHIFT)
#define MV_COMMUTATOR_PRODUCT 			(0x10<<MV_PRODUCT_SHIFT)
#define MV_ANTICOMMUTATOR_PRODUCT		(0x20<<MV_PRODUCT_SHIFT)
#define MV_EUCLIDEAN_PRODUCT				(0x40<<MV_PRODUCT_SHIFT)






#define MV_FLOAT 1	 // use floating point type for reals

#if MV_FLOAT

typedef myfloat Real;
#define Real_SMALL (.0001)	// default gradient epsilon
#define Real_HALF FLOAT_HALF
#define Real_1 FLOAT_1

#endif

// fullbladeindex must be able to hold (1<<MVDIM1)-1 
#if MV_DIM1 <= 8
typedef uint8 fullbladeindex;	
#else
	#if MV_DIM1 <= 32
typedef uint32 fullbladeindex;
	#else
		#if MV_DIM1 <= 64
typedef uint64 fullbladeindex;
		#else 
#pragma error "N > 64 unsupported"
		#endif // 64
 	#endif // 32
#endif // 8

typedef signed char mvgrade;	  // must be able to hold MVDIM1 and -1 to flag maximum
extern mvgrade g_N;
extern mvgrade g_N0;



typedef uint8 bladeindex;	// must be able to hold (1<<MVDIM0)-1 
typedef uint8 mvcapacity;	// must be able to hold 1<<MVDIM0 

typedef fullbladeindex mvgradesflag; 	// (N+1)-bit flag 
typedef flagstype mvproduct; 			// product flags 

#define MV_NUM_CONJUGATION (6)
typedef flagstype mvconjugation; 	// 6-bit flag 

typedef	struct multivectorstruct0 * multivector0;
typedef	struct multivectorstruct * multivector;

typedef flagstype mvoperation;
#define op_NIL (0)
#define MV_DUAL (1<<MV_NUM_CONJUGATION)

class mv {
public:
					mv();
explicit	mv(const Real c);
explicit	mv(const vector3D * v);
					mv(mvoperation o, Real c, const multivector m);
					mv(Real c,const multivector a);
explicit	mv(const multivector a);
					mv(fullbladeindex bi, Real c);
 					mv(const mv & src);						// copy constructor
					~mv();

	void clear(void);
	void setOp(mvoperation);
	void writeEnable(void);
	void setDual(bool);
	bool getDual(void) 	 			const;
	void setConj(mvconjugation c);


	bool isZero(void)					const;
	bool isScalar(void)				const;
	bool pureGrade(int)				const;
	mvconjugation conj(void)	const;
	mvgradesflag grades(void)	const;
	fullbladeindex principleBladeIndex(void) const;
	mv						 principleBlade(void) const;


	Real scalarPart(void)			const;

	int	 numBlades(void)			const;

	
	mv  operator=(mv a);					// assignment

	mv 	operator[](int i) const; 	// grade restriction
	mv 	operator[](int i); 				// grade restriction


//	mv  operator~();							// reverse
	mv  operator!();							// involution
	mv  operator-();							// negation
	
	mv  operator+=(Real b);
	mv  operator+=(mv &b);

	mv  operator-=(Real b);
	mv  operator-=(mv &b);

	mv  operator*=(Real b);	// mult by scalar
	mv  operator*=(mv &b);	// geometric product

	mv  operator/=(Real b);	// divide by scalar
	mv  operator/=(mv &b);	// mult by geometric inverse

	mv  operator>>=(mv &b);	// projection into
	mv  operator<<=(mv &b);	// rejection from


	mv  dual(void);					// dual         ( ie *= mv_i_inv)
	mv  antiDual(void);			// inverse dual ( ie *= mv_i_inv * mv_i^2)
	mv	unit(void);					// scale so that square has unit scalar part
	mv	exp(int n=9);					// exponential

	mv	nrm(mvproduct pro=MV_GEOMETRIC_PRODUCT);
	mv	hrm(void);
	mv	rev(void);
	mv	clf(void);
	mv	ivl(void);
	mv	inv(mvproduct pro=MV_GEOMETRIC_PRODUCT);

 	multivector 	m;
	Real					s;
	
	mvoperation		op;
//	boolean				toFree; 
}; // endclass 

mv scalar(const Real b);
mv blade(const fullbladeindex bi, const real c);
mv versor(Real s,const fullbladeindex bi,int k);

mv factorise(mv a);
mv transform(mv a,mv x);
mv exponentiatedtransform(mv a,mv x,Real theta); // e^(theta*a) x e^(-theta*a)

mv	operator^(mv &a, int i);						// raise to power 

mv  operator+(const mv &a, const Real b);
mv  operator+(const Real b,const mv &a);
mv  operator+(const mv &a, mv & b);
//mv  operator+(const mv &a, mv &b);

mv  operator-(const mv &a, const Real b);
mv  operator-(const Real b,const mv &a);
mv  operator-(const mv &a, mv & b);
//mv  operator-(const mv &a, const mv &b);


mv  operator*(const mv &a, const mv & b); // * is geomeric product
mv  operator*(const mv &a, const Real b);
mv  operator*(const Real b,const mv &a);

mv  operator/(mv &a, mv &b);
mv  operator/(const mv &a,const mv &b);
mv  operator/(const mv &a, const Real b);
mv  operator/(const Real b,mv &a);



mv 	operator&(mv &a,mv &b); 						// & is commutator product
mv 	operator^(const mv &a,const mv &b); // ^ is outter product      (as Gaigen)
mv 	operator|(const mv &a,const mv &b); // | is left contraction    (Gaigen uses << but reserve that for IO)
mv 	operator||(mv &a, mv &b);						// || is join								(Gaigen uses |)
mv 	operator&&(mv &a, mv &b);						// && is meet								(Gaigen uses &)

Real operator%(const mv &a,const mv &b);// % is scalar product      (as Gaigen)

mv mv_delta_product(const mv &a,const mv &b,mvproduct pro=MV_GEOMETRIC_PRODUCT);
mv mv_right_contraction(const mv &a,const mv &b);
mv mv_product(const mv &a,const mv &b,mvproduct pro);

boolean operator<(mv & a,mv & b);

typedef mv (*mv_functionptr)(mv &x,const mv *);

class mvfunc { public:	mvfunc();	mvfunc(mv_functionptr); char * name;	mv_functionptr	func;	mv 	para[3];};


mv RandomOneVec(mvgrade n,Real sq);

mv	spinorform(Real * mag,Real * rapid,mv x,mv a);

mv	OneGradient(mvfunc &f,mv &x,int N=-1,Real epsilon=Real_SMALL);

mv pointswedge(int k,posn3D * p);
mv kplane(int k,posn3D * p);


void point(mv &x, posn3D *	p);


mv 	point(posn3D *	p);
mv 	sphere(posn3D * c,real radius);	 // GHC-1 embedding 
mv 	sphere2(posn3D * c,real radius); // GHC-2 embedding

mv RandomBlade(mvgrade grade,mvgrade n);


void ToVector(vector3D * pv,mv a);


int FactoriseGeometric(mv * tgt,mv a,mvconjugation cnj,int max);
int FactoriseOutter(mv * tgt,mv a,mvconjugation cnj,int max);

extern boolean g_mvfactored;	 // 1=> build factored mv products 
extern mvcapacity g_MVmargin;


extern int g_ghc; // current level of generalised homogenised coords 
// global multivectors 

extern mv mv_basis[MV_DIM1];
extern mv mv_unit_cube[MV_DIM1];

#define MV_NUM_POINT_REG (4)

extern mv mv_p_reg[MV_MAX_GHC][MV_NUM_POINT_REG];

#if MV_MAX_GHC
extern mv mv_E[MV_MAX_GHC];
extern mv mv_Ee123[MV_MAX_GHC];
extern mv mv_ep[MV_MAX_GHC];
extern mv mv_em[MV_MAX_GHC];
extern mv mv_e0[MV_MAX_GHC], mv_e0_inv[MV_MAX_GHC];
extern mv mv_ee[MV_MAX_GHC], mv_ee_inv[MV_MAX_GHC];

extern mv mv_I[];
extern mv mv_I_inv[];
#endif

extern mv mv_i,mv_i_inv;

extern mv mv_0;
extern mv mv_1;

extern mv e1,e2,e3;
extern mv e12,e13,e23,e123;
#if MV_DIM1>=4
	extern mv e4,e124,e1234;
	#if MV_DIM1>=5
		extern mv e5,e345,e12345;
			#if MV_DIM1>=6
				extern mv e6,e123456;
					#if MV_DIM1>=7
						extern mv e7,e1234567;
					#endif // 7
			#endif // 6
	#endif // 5
#endif // 4



void InitMvs(int);
void KillMvs(int);


#define MV_BPF_BASIC 2

#define MV_BPF_GEOMETRIC 4
#define MV_BPF_NORMALISE (0x08)
#define MV_BPF_UNIT_e0  0x10
#define MV_BPF_SEPERATE_EE  0x20
#define MV_BPF_SEPERATE_E0  0x40
#define MV_BPF_SHOWZERO  0x80


#if MV_DEBUG
extern int DiagnoseMV(multivector m);
extern int Diagnosemv(mv m);
#endif


#if WATCOM
#pragma warning 604 9 ;// Must look ahead to determine whether construct is declaration or expression
#pragma warning 595 9 ;//  ...resolved as expression
#pragma warning 665 9 ;//  temporary object used to initialise
#pragma note 638 9 ;//  note for 665
#endif
